#!/usr/bin/perl

print ":: pbnc by Tom Kistner aka duncanthrax\n:: http://duncanthrax.net/pbnc/\n";

use IO::Select;
use IO::Socket;
use Net::Netmask;
use POSIX qw(setsid);

$| = 1;

# -- Read config file --------------------------------------
unless ($ARGV[0]) {
  print "Usage: ./pbnc.pl <CONFIGFILE> [debug]\n";
  exit(0);
};

$cfg = read_ph($ARGV[0]);
unless (ref($cfg) eq 'HASH') {
  print "Config error, aborting.\n";
  exit(1);
};

$debug = 1 if ($ARGV[1] eq 'debug');

$sel = new IO::Select();

$listen_socks = {};
$conn_socks = {};
$conn = {};
$client_netblocks = {};

# -- Open listening sockets --------------------------------
foreach my $site (keys(%{ $cfg })) {

  my ($addr,$port) = split /\:/, $cfg->{$site}->{local};

  $listen_socks->{$site} =
    new IO::Socket::INET(Listen => 1,
                         ReuseAddr => 1,
                         LocalAddr => $addr, 
                         LocalPort => $port);

  if (!defined($listen_socks->{$site})) {
    print "Can't listen on local addr:port $addr:$port, please check configuration\n";
    exit(1);
  };
  
  $sel->add($listen_socks->{$site});  
  
  # add allowed networks IPs for this site
  $client_netblocks->{$site} = [];
  foreach my $netblock (@{ $cfg->{$site}->{client_netblocks} }) {
    my $nn = new2 Net::Netmask($netblock);
    if (defined($nn)) {
      push @{ $client_netblocks->{$site}  }, $nn;
    }
    else {
      print $site.": warning - '$netblock' is not a valid network block, skipping.\n";
    };
  };
  
};

unless ($debug) {
  # -- Launch into background --------------------------------
    print "All OK, detaching from terminal .. ";
    my $pid = fork();
    if (!defined $pid) {
      print "Error:fork() failure.\n";
      exit(0);
    }
    elsif ($pid > 0) {
      print "fork OK [$pid], parent exiting.\n";
      exit(0);
    };
  # ----------------------------------------------------------
  
  # -- Detach from TTY ---------------------------------------
  $0 = "pbnc";
  setsid();
  # ----------------------------------------------------------
};

# -- Set up signal handlers ----------------------------------
$SIG{HUP} = \&quit;
$SIG{TERM} = \&quit;
$SIG{INT} = \&quit;
# ------------------------------------------------------------


# -- Main loop ----------------------------------------------
RESTART:
while(my @ready = $sel->can_read) {

  my $now = time();

  # -- loop through all ready sockets ------------------------------------------
  SOCKET:foreach my $sockref (@ready) {

    # -- accept new connections ------------------------------------------------
    foreach my $site (keys(%{ $cfg })) {
      if ($sockref == $listen_socks->{$site}) {

        print STDERR "accepting new connection for '$site'\n" if ($debug);

        # client connect
        my $client = $listen_socks->{$site}->accept;

        unless (defined($client)) {
          print STDERR "accept failed\n" if ($debug);
          goto RESTART;
        };

        my $conn_id = $client->peerhost().':'.$client->peerport().'->'.$site;
        print STDERR $conn_id.": connecting to server.\n" if ($debug);

        my $target = $cfg->{$site}->{decoy} || '';
        # check if the client is allowed to connect to the real site
        foreach my $nn (@{ $client_netblocks->{$site} }) {
          if ($nn->match($client->peerhost())) {
            $target = $cfg->{$site}->{target};
            last;
          };
        };

        unless ($target) {
          # no target and no decoy
          print STDERR "client is not allowed to connect to $site, and no decoy given.\n" if ($debug);
          eval {
            $client->close();
          };
          goto RESTART;
        };

        # connect to server
        my ($addr,$port) = split /\:/, $target;
        my $server = 
          new IO::Socket::INET(Timeout => 2,
                               PeerAddr => $addr,
                               PeerPort => $port);
        unless ($server) {
          print STDERR "connection to remote server failed\n" if ($debug);
          eval {
            $client->close();
          };
          goto RESTART;
        };

        print STDERR $conn_id.": connection now active.\n" if ($debug);

        # add sockets to select checks
        $sel->add($client);
        $sel->add($server);

        # add sockets to connection hash
        $conn->{$conn_id}->{client} = $client;
        $conn->{$conn_id}->{server} = $server;
        $conn->{$conn_id}->{'last'} = $now;
        $conn->{$conn_id}->{'site'} = $site;

        # add reverse mapping from sockets to conn_id
        $conn_socks->{$client} = [ 'client', $conn_id ];
        $conn_socks->{$server} = [ 'server', $conn_id ];
   
        # we must leave here so we do not
        # treat the listen socket with
        # the data copy routines below.
        goto RESTART;
      };
    };
    # ----------------------------------------------------------------------------  

    # -- copy data between connections -------------------------------------------
    my $tmp = '';
    my ($type, $conn_id) = @{ $conn_socks->{$sockref} };
    if (defined($conn->{$conn_id}->{$type}->recv($tmp, 2048))) {
      if (length($tmp) > 0) {
        # write data to other socket
        my $target = 'client';
        $target = 'server' if ($type eq 'client');
        # check connected state
        if (defined($conn->{$conn_id}->{$target}->connected())) {
          eval {
            $conn->{$conn_id}->{$target}->send($tmp);
          };
          $conn->{$conn_id}->{'last'} = $now;
        }
        else {
          print STDERR $conn_id.": ".$target." connection is dead.\n" if ($debug);
          nuke_conn($conn_id);
          goto RESTART;
        };
      }
      else {
        print STDERR $conn_id.": ".$type." read OK, but no data.\n" if ($debug);
        nuke_conn($conn_id);
        goto RESTART;
      };
    }
    else {
      print STDERR $conn_id.": ".$type." read failed.\n" if ($debug);
      nuke_conn($conn_id);
      goto RESTART;
    };
  }

  # -- Nuke stale connections ----------------------------------------------------
  foreach my $conn_id (keys %{ $conn }) {
    if ($conn->{$conn_id}->{'last'} < ($now - $cfg->{$conn->{$conn_id}->{site}}->{idle_timeout})) {
      print STDERR $conn_id.": idle timeout.\n" if ($debug);
      nuke_conn($conn_id);
    };
  };
}

sub nuke_conn {
  my $conn_id = shift;
  
  $sel->remove($conn->{$conn_id}->{client});
  $sel->remove($conn->{$conn_id}->{server});
  
  my $clientref = $conn->{$conn_id}->{client};
  my $serverref = $conn->{$conn_id}->{server};
  
  eval {
    $conn->{$conn_id}->{client}->close();
    $conn->{$conn_id}->{server}->close();
  };
  
  delete $conn_socks->{$clientref};
  delete $conn_socks->{$serverref}; 
  delete $conn->{$conn_id};
};


sub quit {
  
  # properly close sockets before we close
  foreach my $site (keys %{ $listen_socks }) {
    print STDERR $site.": closing listen socket\n" if ($debug);
    eval {
      $listen_socks->{$site}->close();
    };
  };
  
  foreach my $conn_id (keys %{ $conn }) {
    eval {
      $conn->{$conn_id}->{client}->close();
      $conn->{$conn_id}->{server}->close();
    };
  };
  
  exit(0);
};

sub read_ph {
  my $file = shift;
  open(PH,"< $file");
  undef $/;
  my $tmp = (eval(<PH>));
  print STDERR "Eval Error: ".$@."\n" if ($@);
  $/ = "\n";
  close(PH);
  return $tmp;
};

